#!/usr/bin/sage
#
# SageMath script for checking unisolvence.
#
# $Id: quadrule-ml-symbolic-check.sage,v 1.18 2016/12/07 13:46:46 zlb Exp $
#

def usage():
    if len(sys.argv) > 1:
	print "Invalid command line arguments:", sys.argv[1:]
    print 'Usage:'
    print '    sage', sys.argv[0], 'dim order [order_1 ... order_dim]'
    print 'Note: dim must be in the range [1:35).'
    exit()

if len(sys.argv) <= 1:
    usage()
if (len(sys.argv) != 3 and len(sys.argv) != 3+int(sys.argv[1])) or \
	not sys.argv[1].isdigit() or not int(sys.argv[1]) in range(0,35):
    usage()

dim = int(sys.argv[1])
orders = []
order = 1
for a in sys.argv[2:]:
    if not a.isdigit() or int(a) <= 0 or int(a) < order:
	usage()
    order = int(a)
    orders += [order]

if os.getenv('NUMERIC', 'True').lower() == 'true':
    numeric = True
    print "Using random fractional numbers."
else:
    numeric = False
    print "Using full symbolic evaluation."

import random

#==================== The variables for the abscissas
ratpoly.<va0,va1,va2,va3,va4,va5,va6,va7,va8,va9,va10,va11,va12> = QQ[]
va_list=[va0,va1,va2,va3,va4,va5,va6,va7,va8,va9,va10,va11,va12]

#----- The barycentric coordinates functions
def lam(coor, i):
    return coor[i]

#----- Function converting an integer to a base 36 string
def base36(n):
    if n < 10:
	return chr(n+ord('0'))
    elif n < 36:
	return chr(n-10+ord('a'))
    else:
	print "Error in base36(): invalid argument:", n
	exit()

#----- Function returning the list of all permutation stars for given dim
def get_stars(dim):
    def get_star0(dim, ub, star):
	if dim == 0:
	    res.append(star)
	    return
	for d in range(ub, 0, -1):
	    ub0 = dim - d
	    if ub0 > d:
		ub0 = d
	    get_star0(dim-d, ub0, star+base36(d))
    res = []
    get_star0(dim+1, dim+1, "")
    return res

#----- Function returning the orbit pattern of the uniform grid
def uniform_pattern(dim, orders):
    def recur(dim, order, ub, work):
	if order == 0 or dim == 0:
	    if order > ub:
		return
	    elif order > 0:
		orbits.append(work+[order])
	    return
	for i in range(min(ub,order), 0, -1):
	    recur(dim-1, order-i, i, work+[i])
    # First, get list of orbits
    orbits = []
    for d in range(dim+1):
	if len(orders) == dim+1:
	    order = orders[d]
	elif len(orders) == 1:
	    order = orders[0]
	else:
	    print 'uniform_pattern(): invalid orders[] array'
	    exit()
	recur(d, order, order, [])
    res = {s:0 for s in get_stars(dim)}
    for o in orbits:
	# Determine permutation star of the orbit o
	if len(o) < dim+1:
	    o += [0] * (dim + 1 - len(o))
	tmp = [0] * (order + 1)
	for c in o:
	    tmp[c] += 1
	s = ""
	tmp.sort(reverse=True)
	for t in tmp:
	    if t == 0:
		 break
	    s += base36(t)
	# Increment the counter of the star
	res[s] += 1
    # delete 0 entries
    for k in res.keys():
	if res[k] == 0:
	    del res[k]
    return res

#----- Function returning the list of functions of a basis
def basis(dim, orders, coor):
    def basis_recur(dim, order, coor, I, B):
	if order <= 0:
	    return
	if dim == 0:
	    bas.append(B * lam(coor,I[0])^order)
	    return
	for o in range(1,order+1):
	    if o == 0:
		b = B
	    else:
		b = B * lam(coor,I[0])^o
	    basis_recur(dim-1, order-o, coor, I[1:], b)
    bas = []
    for s in subsets(set([i for i in range(dim+1)])):
	d = len(s) - 1
	if d >= 0:
	    if len(orders) == dim+1:
		o = orders[d]
	    elif len(orders) == 1:
		o = orders[0]
	    else:
		print 'basis(): invalid orders[] array'
		exit()
	    basis_recur(d, o, coor, [i for i in s], 1)
    return bas

#----- Function returning the list of points of a symmetry orbit
# Note: 'vars' provides the len(star)-1 free abscissas and the last abscissae
# is computed such that the sum of all abscissas equals 1.
#
# If len(vars)<len(star)-1, then the missing abscissas will be set to
# random fractional numbers.
def perm(star, vars):
    def perm0(star, vars, lam):
	if star == "1":
	    orbit.append(lam+[vars[0]])
	    return
	for i, c in enumerate(star):
	    if c == '1':
		if i == 0:
		    s = star[i+1:]
		    v = vars[i+1:]
		else:
		    s = star[0:i] + star[i+1:]
		    v = vars[0:i] + vars[i+1:]
	    else:
		if i == 0:
		    s = base36(int(c,36)-1) + star[i+1:]
		else:
		    s = star[0:i] + base36(int(c,36)-1) + star[i+1:]
		v = vars
	    perm0(s, v, lam+[vars[i]])
    orbit = []
    sum = 1
    for i in range(min(len(vars),len(star)-1)):
	if numeric:
	    vars[i] = random.randint(0,int(sum*1000000/int(star[i],36)))/1000000
	sum -= int(star[i],36) * vars[i]
    # assign random fractional values to missing variables
    for i in range(len(vars), len(star)-1):
	vars += [random.randint(0,int(sum*1000000/int(star[i],36)))/1000000]
	sum -= int(star[i],36) * vars[i]
    # generate recursively the orbit list
    perm0(star, vars[:len(star)-1] + [sum/int(star[-1:],36)], [])
    return orbit

# Function checking unisolvence of the point set P w.r.t. basis of orders[]
def unisolvence(dim, orders, P):
    A = matrix([basis(dim, orders, p) for p in P])
    n = A.nrows()
    m = A.ncols()
    #print 'm = ' + str(m) + ', n = ' + str(n)
    if n != m:
    	print "Error: m and n mismatch."
    if not numeric:
	B = A.echelon_form('frac',include_zero_rows=True)
    else:
	B = A.echelon_form(include_zero_rows=True)
    res=1
    for diag in B.diagonal():
	if diag == 0:
	    res = 0
    return res != 0

#----- Function returning the list of the solutions to the Diophantine equation
def Diophantine(dim, orders):
    def evaluate(dim, orders, stars):
	P = []
	va_start = 0
	for s in stars:
	    n = len(s)-1
	    for i in range(stars[s]):
		if not numeric:
		    if va_start+n > len(va_list):
			print "Error: insufficient number of variables, abort."
			print "stars =", stars
			exit()
		    vars = va_list[va_start:va_start+n]
		    va_start += n
		else:
		    vars = []
		P += perm(s, vars)
	res = unisolvence(dim, orders, P)
	if stars == stars_uniform:
	    prefix = ' ***'
	else:
	    prefix = ' ---'
	if res:
	    print prefix, "[01;31mTrue [0m", str(stars).replace(': ', ':')
	else:
	    print prefix, "[01;34mFalse[0m", str(stars).replace(': ', ':')
	if (stars == stars_uniform) != res:
	    print "Failed case found, stop."
	    exit()
	return

    def solve(eqn, n, sol):
	if n <= 0:
	    if n == 0:
		evaluate(dim, orders, sol)
	    return
	if eqn == {}:
	    return
	d = gcd(eqn.values())
	if n % d != 0:
	    return
	k = eqn.keys()[0]
	eqn1 = eqn.copy()
	del eqn1[k]
	for i in range(int(n/eqn[k])+1):
	    sol1 = sol.copy()
	    if i > 0:
		sol1[k] = i
	    solve(eqn1, n - i * eqn[k], sol1)

    if len(orders) != dim+1 and len(orders) != 1:
	print 'Diophantine(): invalid orders[] array.'
	exit()
    stars = get_stars(dim)
    eqn = {s : len(perm(s,[])) for s in stars}
    n = 0
    for d in range(dim+1):
	if d == 0:
	    n += dim + 1
	else:
	    if len(orders) == 1:
		order = orders[0]
	    else:
		order = orders[d]
	    if order > d:
		p = 1; q = 1
		for k in range(1,d+1):
		    p *= order - k; q *= k
		n += (p/q) * factorial(dim+1)/(factorial(d+1)*factorial(dim-d))
    #print 'The Diophantine equation:', eqn, '=', n
    stars_uniform = uniform_pattern(dim, orders)
    del eqn[base36(dim+1)]
    solve(eqn, n, {})
    solve(eqn, n-1, {base36(dim+1):1})

Diophantine(dim, orders)
